// Copyright (c) 2016--19, Norman Gray
// All rights reserved.
// 
// Redistribution and use in source and binary forms,
// with or without modification,
// are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

// --------
// unittests module, revision c16d1b5a65ac, of 2020-06-25 19:05 +0100
// See https://code.nxg.name/nxg/unittests
// --------


// Simple unit testing support

// To use, create a test file test-foo.cpp:
//
// #include "unittests.h"
//
// void test_foo(void)
// {
//     /* stuff... */
//     assert_*(xxx)
// }
//
// int main(int argc, char** argv)
// {
//     Testcase funcs[] = {
//         { "description", test_foo, },
//     };
//     int nfuncs = sizeof(funcs)/sizeof(funcs[0]);
//
//     return run_test_cases("suite-name", funcs, nfuncs, argc, argv);
// }
//
// Assertions can be:
//
//   assert_equal(expected, actual);  // various types
//   assert_equal(expected, actual, tolerance); // ...for floats
//   assert_true(actual, msg);        // argument is non-zero
//   assert_false(actual, msg)        // argument is zero
//   assert_null(actual, msg);        // argument is NULL
//   assert_not_null(actual, msg);    // argument is not NULL
//   assert_fail(msg);                // unconditional fail, with message
//   SHOULD_FAIL(msg, funcname, arg1, arg2, ...); // check throws C++ exception
//
// Build with
//
//   cxx -o test-foo.o -c test-foo.cpp
//   cxx -o unittests.o -c unittests.cpp
//   cxx -o test-foo test-foo.o unittests.o
//   ./test-foo            # run all the tests
//   ./test-foo descrip    # run the tests which have 'descrip' in their description
//   ./test-foo -h         # show options
//
// The tests can also be run on an Arduino.  Invoke as something like:
//
//     Testcase funcs[] = {
//         // as above..
//     };
//     int nfuncs = sizeof(funcs)/sizeof(funcs[0]);
//
//     const char* dummy_argv[] = { "test-hhc-arduino", "-v" };
//     size_t nargs = sizeof(dummy_argv)/sizeof(dummy_argv[0]);
//
//     return run_test_cases(NULL, funcs, nfuncs, nargs, dummy_argv);
//
// You can also compile with (some suitable equivalent of) -DTestSerial=Serial1
// to have chatter sent to a different serial port.


#if ARDUINO
#include <Arduino.h>
#else
#include <iostream>
#include <sstream>
#endif

#include <stdarg.h>
#include <math.h>
#include <string.h>

#include "unittests.h"

#define MSGLEN 256

#if ARDUINO
#include <setjmp.h>
static jmp_buf unittest_jmp_buf_;
static char error_msg[MSGLEN];
#endif

static const char* _testname = "TESTNAME";

#if ARDUINO

// Compile with (eg) -DTestSerial=Serial1 if you want chattering to go to a different serial port
#ifndef TestSerial
#define TestSerial Serial
#endif

class SimpleStreamTrait {};
SimpleStreamTrait endl;
class SimpleStream {
public:
    SimpleStream(void) {};
    SimpleStream& operator<<(const char* const);
    SimpleStream& operator<<(const char);
    SimpleStream& operator<<(const int);
    SimpleStream& operator<<(SimpleStreamTrait&);
};

SimpleStream& SimpleStream::operator<<(const char* const s)
{
    TestSerial.print(s);
    return *this;
}

SimpleStream& SimpleStream::operator<<(const char c)
{
    TestSerial.print(c);
    return *this;
}

SimpleStream& SimpleStream::operator<<(const int i)
{
    TestSerial.print(i);
    return *this;
}

SimpleStream& SimpleStream::operator<<(SimpleStreamTrait& unused)
{
    (void)unused;            // suppress 'unused variable' warning
    TestSerial.println();
    return *this;
}
SimpleStream cout;
#define cerr cout

#else
using std::cout;
using std::cerr;
using std::endl;
#endif

// If fail_on_error_p is false, then we just keep on ploughing through
// (ie, it's for debugging)
bool fail_on_error_p = true;

void assert_fail_(const char* fname, const int line, const char* fmt, ...)
{
#if !ARDUINO
    static char error_msg[MSGLEN];
#endif

    size_t ofs = snprintf(error_msg, MSGLEN, "test %s: ", _testname);
    if (ofs >= MSGLEN) {
        // ye gods -- how long was that _testname?!
        snprintf(error_msg, MSGLEN, "test name ridiculously long!");
    } else {
        va_list ap;
        va_start(ap, fmt);
        ofs += vsnprintf(error_msg+ofs, MSGLEN-ofs, fmt, ap);
        va_end(ap);

        if (ofs < MSGLEN-ofs) {
            snprintf(error_msg+ofs, MSGLEN-ofs, " (%s:%d)", fname, line);
        }
    }

    if (fail_on_error_p) {
#if ARDUINO
        longjmp(unittest_jmp_buf_, 1);
#else
        throw error_msg;
#endif
    } else {
        cerr << error_msg << endl;
    }
}

void set_test_name(const char* name)
{
    _testname = name;
}

#if ARDUINO
const char* get_test_name(void)
{
    return _testname;
}

#else
std::string get_test_name(void)
{
    return std::string(_testname);
}
#endif

static int _test_count = 0;
static int get_test_count(void)
{
    int rval = _test_count;
    _test_count = 0;
    return rval;
}

static int verbosity = 1;
bool get_test_verbose_p(void)
{
    return verbosity > 1;
}

int set_test_verbosity(int newval)
{
    int oldval = verbosity;
    verbosity = newval;
    return oldval;
}

#if !ARDUINO
// Assert that two strings are equal, and throw a string exception if not
void assert_equal_(const char* fname, const int line,
                   const std::string expected, const std::string actual)
{
    _test_count++;
    if (actual.compare(expected) == 0)
        return;
    else
        assert_fail_(fname, line, "expected string <%s>, got string <%s>", expected.c_str(), actual.c_str());
}

// the same, but with a char* expected string
void assert_equal_(const char* fname, const int line,
                   const char* expected, const std::string actual)
{
    _test_count++;
    if (actual.compare(expected) == 0)
        return;
    else
        assert_fail_(fname, line, "expected char* <%s>, got string <%s>", expected, actual.c_str());
}
#endif

// the same, but comparing pointer values
void assert_equal_(const char* fname, const int line, const void* expected, const void* actual)
{
    _test_count++;
    if (expected == actual) return; // JUMP OUT

    assert_fail_(fname, line, "expected void* <%p>, got <%p>", expected, actual);
}

void assert_equal_(const char* fname, const int line,
                   const char* expected, const char* actual)
{
    _test_count++;
    while (*expected == *actual) {
        if (*expected == '\0') {
            // actually the end of the string
            return;
        }
        expected++;
        actual++;
    }
    assert_fail_(fname, line, "expected char* <...%s>, got char* <...%s>", expected, actual);
}

void assert_equal_(const char* fname, const int line,
                   const long expected, const long actual)
{
    _test_count++;
    if (expected == actual) return; // JUMP OUT

    assert_fail_(fname, line, "expected long <%ld>, got <%ld>", expected, actual);
}

// check that expected = actual, to within the given fractional tolerance;
// if, however, expected is precisely zero, then check actual is
// within the absolute range (-tolerance, tolerance).
// If _tolerance_ is precisely zero, then check that actual and expected are precisely equal.
void assert_equal_(const char* fname, const int line,
                   const double expected, const double actual, const double tolerance)
{
    _test_count++;

    if (tolerance == 0.0) {
        if (actual == expected) return; // JUMP OUT
        assert_fail_(fname, line, "expected double %g, got %g (absolute)",
                     expected, actual);

    } else if (expected == 0.0) {
        // special case
        if (fabs(actual) <= tolerance) return; // JUMP OUT
        assert_fail_(fname, line, "expected zero, got %g (%g > %g, absolute range)",
                     actual, fabs(actual), tolerance);

    } else {
        if (fabs(actual / expected - 1) <= tolerance) return; // JUMP OUT

        assert_fail_(fname, line, "expected double %g, got %g (%g > %g)",
                     expected, actual, fabs(actual / expected - 1), tolerance);
    }
}

// check that expected = actual, to within the given _absolute_ tolerance
void assert_equal_abs_(const char* fname, const int line,
                       const double expected, const double actual, const double tolerance)
{
    _test_count++;
    if (fabs(actual - expected) < tolerance) return; // JUMP OUT

    // if we get here, we've failed the test
    assert_fail_(fname, line, "expected double %g, got %g (%g > %g, absolute)",
                 expected, actual, actual - expected, tolerance);
}

void assert_true_(const char* fname, const int line,
                  const int t, const char* msg)
{
    _test_count++;
    if (t) return;              // JUMP OUT

    assert_fail_(fname, line, "(%s) expected true, but wasn't", msg);
}

void assert_false_(const char* fname, const int line,
                  const int t, const char* msg)
{
    _test_count++;
    if (! t) return;            // JUMP OUT

    assert_fail_(fname, line, "(%s) expected false, but wasn't", msg);
}

void assert_null_(const char* fname, const int line,
                  const void* p, const char* msg)
{
    _test_count++;
    if (p == NULL) return;      // JUMP OUT

    assert_fail_(fname, line, msg);
}

void assert_not_null_(const char* fname, const int line,
                     const void* p, const char* msg)
{
    _test_count++;
    if (p != NULL) return;      // JUMP OUT

    assert_fail_(fname, line, msg);
}

void assert_pass(void)
{
    _test_count++;
}

#if ARDUINO
static void Usage(const char* progname)
{
    cerr << "Bad call to run_test_cases! in " << progname << endl;
}
#else
static void Usage(const char* progname)
{
    cerr << "Usage: " << progname << " [-lhqv] [selector]" << endl;
    cerr << "Run tests:" << endl;
    cerr << "with -v or environment VERBOSE=2, run verbosely" << endl;
    cerr << "with -q or environment VERBOSE=0, run silently" << endl;
    cerr << "with -l, only list the available test names" << endl;
    cerr << "with -f, force through the tests -- don't exit on failures" << endl;
    cerr << "If a 'selector' is present, then run only tests containing that substring" << endl;
    exit(1);
}
#endif

#if ARDUINO
// Wrap the setjmp and possibly longjmp in a separate function, to
// avoid (largely spurious in practice) warnings about automatic
// variables being clobbered.  This also further isolates the
// irreducible setjmp/longjmp nastiness.
//
// There's a good explanation at <https://stackoverflow.com/questions/2024933/>
bool call_ok(void (*f)(void))
{
    bool called_ok;

    if (setjmp(unittest_jmp_buf_)) {
        called_ok = false;
    } else {
        (*f)();
        called_ok = true;
    }
    return called_ok;
}
#endif

// Run a sequence of test cases.
// The argc and argv are from the calling main program, and provide
// options and possible test selectors.
//
// Typically, argc > 0, but if it is zero (which is more convenient
// when we're calling from an Arduino), that's handled.
// The argument suite_name may be null (in which case it is defaulted from argv).
int run_test_cases(const char* suite_name,
                   const Testcase* cases, const size_t n,
                   int argc, const char* const *argv)
{
    int nok = 0;
    int nfail = 0;

    bool list_only = false;

    const char* progname;
    if (argc == 0) {
        // this will happen only in the Arduino case
        progname = "UNITTESTS";
    } else {
        progname = argv[0];
        argc--;
        argv++;
    }

    if (suite_name == nullptr) {
        // make suite_name be the last element in the progname path
        suite_name = progname;
        while (*suite_name != '\0') suite_name++;
        while (*suite_name != '/' && suite_name > progname) suite_name--;
        if (suite_name != progname) {
            // ie, there was a slash in the path
            suite_name++;       // point after the slash
        }
    }

    // We do the option processing by hand, rather than using getopt,
    // because (i) the option handling is simple, (ii) we therefore
    // avoid a separate dependency, but most importantly because (iii)
    // this allows us to declare the argument argv to be a pointer to
    // const char (the argv argument in POSIX getopt is `char * const argv[]`).
    for (/* nothing */; argc > 0; argc--, argv++) {

        if (**argv != '-') break; // end of options

        for (const char* p = &argv[0][1];
             *p != '\0';
             p++) {
            switch (*p) {
              case 'l':
                list_only = true;
                break;
              case 'q':
                set_test_verbosity(0);
                break;
              case 'v':
                set_test_verbosity(2);
                break;
              case 'f':
                fail_on_error_p = false;
                break;
              case 'h':             // FALLTHROUGH
              default:
                Usage(progname);
            }
        }
    }

    const char* const selector = (argc > 0 ? argv[0] : nullptr);

    char* verbose_env = getenv("VERBOSE");
    if (verbose_env != NULL) {
        int verbose_flag = strtol(verbose_env, NULL, 10);
        set_test_verbosity(verbose_flag);
    }

    if (list_only) {
        cout << "Available tests:" << endl;
        for (size_t i=0; i<n; i++) {
            cout << "  " << cases[i].name << endl;
        }
        nfail = 0;
    } else {
        for (size_t i=0; i<n; i++) {
            if (selector != NULL && strstr(cases[i].name, selector) == NULL) {
                if (verbosity > 1) {
                    cerr << "[skipping test: " << cases[i].name << ']' << endl;
                }
                continue;
            }
            if (verbosity > 1) {
                cerr << "======== Testing " << suite_name << '/' << cases[i].name
                     << "..." << endl;
            }

            set_test_name(cases[i].name);
#if ARDUINO
            if (call_ok(cases[i].f)) {
                // normal case
                int ntests = get_test_count();
                if (verbosity > 1) {
                    cerr << ntests << " ok" << endl;
                }
                nok += ntests;
            } else {
                // error
                cerr << cases[i].name << ": test failed: " << error_msg << endl;
                nfail++;
            }
#else
            try {
                (*cases[i].f)();
                int ntests = get_test_count();
                if (verbosity > 1) {
                    cerr << ntests << " ok" << endl;
                }
                nok += ntests;
            } catch (const char* e) {
                nfail++;
                cerr << cases[i].name << ": test failed with (string) exception: " << e << endl;
            } catch (std::exception& e) {
                nfail++;
                cerr << cases[i].name << ": test failed with exception: " << e.what() << endl;
            } catch (std::exception* e) {
                nfail++;
                cerr << cases[i].name << ": test failed with exception: " << e->what() << endl;
            } catch (...) {
                // eh?
                nfail++;
                cerr << cases[i].name << ": unexpected exception!" << endl;
            }
#endif
        }

        if (verbosity > 0) {
            if (nfail == 0) {
                cerr << "OK (" << suite_name << "):" << nok << " tests" << endl;
            } else {
                cerr << "FAIL (" << suite_name << "):" << nfail << '/' << nfail+nok << endl;
            }
        }
    }

    return nfail;
}
